(* 
   Preuves d'egalites dans les anneaux integres.
   Tactique Gb
   Jerome Creci et Loic Pottier
   novembre 2003
*)

Declare ML Module "utile".
Declare ML Module "entiers".
Declare ML Module "polynomes2".
Declare ML Module "dansideal".
Declare ML Module "gbcoq".

Require Export Reals.
Require Export ZArith.
Require Export Classical_Prop.
Require Export Fourier.

Section Domains.

Variable A : Type.

Variable Aplus : A -> A -> A.
Variable Amult : A -> A -> A.
Variable Aone : A.
Variable Azero : A.
Variable Aopp : A -> A.
Variable Aeq : A -> A -> bool.

Variable TA:(Ring_Theory Aplus Amult Aone Azero Aopp Aeq).

Add Abstract Ring A Aplus Amult Aone Azero Aopp Aeq TA.

Infix 4 "+" Aplus.
Infix 4 "*" Amult.
Notation "0" := Azero.
Notation "1" := Aone.
Notation "- x" := (Aopp x) (at level 3).

Lemma Th_minus_eq : (n,m:A) n+(-m)==0 -> n==m.
Intros.
Replace n with n+-m+m.
Rewrite -> H; Ring.

Ring.
Qed.

Definition is_domain:=(n,m:A) n*m==0 -> n==0 \/ m==0.

Lemma Th_integrite :is_domain -> (n,m:A) n*m==0 -> n==0 \/ m==0.
Unfold is_domain.
Auto.
Qed.

Lemma Th_mult_cte :is_domain ->(a,n:A) ~(a==0) -> a*n==0 -> n==0. 
Unfold is_domain.
Intros.
Elim (H ? ? H1).
Intuition.

Auto.
Qed.

End Domains.

Lemma R_is_a_domain:(is_domain R Rmult R0).
Red.
Intros.
Apply without_div_Od.
Auto.
Qed.
Lemma eq_eqT:(A:Set)(x,y:A)x=y->x==y.
Intros.
Rewrite -> H; Auto.
Qed.

Lemma eqT_eq:(A:Set)(x,y:A)x==y->x=y.
Intros.
Rewrite -> H; Auto.
Qed.

Lemma neq_neqT:(A:Set)(x,y:A)~x=y->~x==y.
Intuition.
Apply H.
Rewrite -> H0; Auto.
Qed.

Lemma Z_is_a_domain:(is_domain Z Zmult ZERO).
Red.
Intros.
Generalize (!Zmult_zero n m).
Intros.
Assert (eq Z (Zmult n m) ZERO).
Apply eqT_eq; Auto.

Elim (H0 H1).
Intros.
Left.
Apply eq_eqT; Auto.

Intros.
Right.
Apply eq_eqT; Auto.
Qed.

Recursive Tactic Definition Gb2 :=
  Match Context With
   |[h:?1=?2|-?]->Generalize (eq_eqT ? ?1 ?2 h);Clear h;Intro h; Try Gb2
   |[|-?1=?2]->Apply eqT_eq; Try Gb2
.

Recursive Tactic Definition Gb1 :=
  Match Context With
     |[h1:(is_domain ?1 ?2 ?3);h:(?4 ?5 ?6)==?7|-?6==?7] ->
             Generalize (Th_integrite ?1 ?2 ?3 h1 ?5 ?6 h);Clear h;Intro h;Elim h;Intros;Try Auto; Try Gb1
.

Tactic Definition Gb :=
   Try Gb2; Gb0 ;  Try Auto; Try DiscrR;
   Try Apply R_is_a_domain;Try Apply Z_is_a_domain;
   Unfold is_domain; Try Gb1; Try Apply neq_neqT;Try Omega
.

