(*i camlp4deps: "parsing/grammar.cma" i*)

(* La tactique Gb *)
open Ast;;
open Auto;;
open Clenv;;
open Command;;
open Constrextern;;
open Coqast;;
open Declarations;;
open Declare;;
open Eauto;;
open Environ;;
open Equality;;
open Evd;;
open Genarg
open Inductive;;
open Inductiveops
open Libnames
open Nameops
open Names;;
open Pfedit;;
open Pp;;
open Printer;;
open Proof_trees;;
open Proof_type;;
open Rawterm
open Reduction;;
open Reductionops
open Refiner;;
open Showproof_ct
open Sign;;
open System
open Tacexpr
open Tacmach;;
open Tacred;;
open Tacticals;;
open Tactics;;
open Term;;
open Termops
open Topconstr;;
open Translate
open Typing;;
open Univ
open Util;;
open Vernacentries;;
open Vernacinterp;;
open Ring;;
open Coqlib;;
open Tauto;;

open Utile;;
open Polynomes2;;

(*****************************************************************************)
(* divers *)

let prt s = 
  print_string (s^"\n");flush(stdout)
;;

let parse_ast = 
  Pcoq.parse_string Pcoq.Constr.constr
;;

(* chaine en syntaxe Coq -> terme des constructions *)
let parse s =  
  Constrintern.interp_constr Evd.empty (Global.env()) (parse_ast s)
;;

(* meme chose dans le contexte d'un but *)
let pf_parse_constr gl s =
  Constrintern.interp_constr Evd.empty (pf_env gl) (parse_ast s)
;;

(* impression d'un terme des constructions *)
let ppterm = 
  (fun x -> pp(prterm x))
;;

let prast c = 
  pp(print_ast c)
;;

let string_of_sp sp = 
  string_of_id (basename sp)
;;

let rec string_of_constr c =
  match kind_of_term c with
      Cast (c,t) -> string_of_constr c
    |Const c -> string_of_kn c
    |Var(c) -> string_of_id c
    |Ind (sp,i) -> string_of_kn sp
    | _ -> "not_of_constant"
;;

(**********************************************************************)
(* oprations sur les fractions.
   Une fraction est un couple d'entiers (bigint) *)

let opp_frac (n,d) = 
  (opp_coef n,d)
;;

let plus_frac (n1,d1) (n2,d2) =
  (plus_coef (mult_coef n1 d2) (mult_coef d1 n2),
   mult_coef d1 d2)
;;

let sub_frac x y = 
  plus_frac x (opp_frac y)
;;

let mult_frac (n1,d1) (n2,d2) =
  (mult_coef n1 n2, mult_coef d1 d2)
;;

let div_frac (n1,d1) (n2,d2) =
  (mult_coef n1 d2, mult_coef d1 n2)
;;

let inv_frac (n1,d1) = (d1,n1)
;;

let frac0 = 
  (coef0,coef1)
;;

let frac1 = 
  (coef1,coef1)
;;

let norm_frac (n,d) =
  let a = pgcd n d in
    (div_coef n a, div_coef d a)
;;

let string_of_constant kn = 
  let l = label kn in
    string_of_label l
;;

(* terme des constructions -> fraction *)
let rec fraction_of_constr c =
  match kind_of_term c with
    | Cast (c,t) -> (fraction_of_constr c)
    | App (c,args) ->
        (match kind_of_term c with
             Const c ->
               (match (string_of_constant c) with
                    "Ropp" -> 
                      opp_frac (fraction_of_constr args.(0))
                  |"Rinv" -> 
                     inv_frac (fraction_of_constr args.(0))
                  |"Rmult" -> 
                     mult_frac (fraction_of_constr args.(0))
                      (fraction_of_constr args.(1))
                  |"Rdiv" -> 
                     div_frac (fraction_of_constr args.(0))
                      (fraction_of_constr args.(1))
                  |"Rplus" -> 
                     plus_frac (fraction_of_constr args.(0))
                      (fraction_of_constr args.(1))
                  |"Rminus" -> 
                     sub_frac (fraction_of_constr args.(0))
                      (fraction_of_constr args.(1))
                  |"Zopp" ->
                     opp_frac (fraction_of_constr args.(0))
                  |"Zmult" -> 
                     mult_frac (fraction_of_constr args.(0))
                      (fraction_of_constr args.(1))
                  |"Zplus" -> 
                     plus_frac (fraction_of_constr args.(0))
                      (fraction_of_constr args.(1))
                  |"Zminus" -> 
                     sub_frac (fraction_of_constr args.(0))
                      (fraction_of_constr args.(1))
                  |"Zs" ->
                     plus_frac (fraction_of_constr args.(0))
                      frac1
                  | _ -> failwith "not an integer")
           |Construct c ->
              (* POS=(("Z",0),2)   
		 NEG=(("Z",0),3)  
		 xI= (("positive",0),1) 
		 x0= (("positive",0),2) *)
              (match (string_of_constant (fst (fst c))) with 
                   "Z" ->
                     (match (snd c) with 
                          2 -> fraction_of_constr args.(0)
                        | 3 -> opp_frac (fraction_of_constr args.(0))
                        |_ -> assert false)
                 |"positive" ->
                    (match (snd c) with 
                         1 -> plus_frac (mult_frac (fraction_of_constr args.(0)) ((coef_of_int 2),coef1)) frac1
                       |2 -> mult_frac (fraction_of_constr args.(0)) ((coef_of_int 2),coef1)
                       |_ -> assert false)
                 |"nat" ->
                    (match (snd c) with 
                         2 -> plus_frac (fraction_of_constr args.(0)) frac1
                       |_ -> assert false)
                 |_ -> assert false)
           |_ -> assert false) 
    | Const c ->
        (match (string_of_constant c) with
             "R1" -> frac1
           |"R0" -> frac0
           |_-> assert false)
    | Construct c ->
        (match (string_of_constant (fst (fst c))) with
             "Z" ->
               (match (snd c) with   
                    1 -> frac0  (* ZERO *)
                  |_ -> assert false)
           |"positive" ->
              (match (snd c) with   
                   3 -> frac1  (* xH *)
                 |_ -> assert false)
           |"nat" ->
              (match (snd c) with
                   1 -> frac0 (* O *)
                 |_-> assert false)
           |_ -> assert false)
    |_-> assert false
;;

(***********************************************************************)
(* Polynomes a coefficients fractionnaires *)
(* un polynome est represent'e par un polynome `a coefficient entiers 
   et un entier (par lequel on divise tous les coefficients *)

let all_vars = ref (Hashtbl.create 51);;
let nvars = ref 0;;
let special_list_vars = ref [];;

type polyf = coef * poly (* denominateur, polynome *)
;;

(* normalise c et les coef de p, rend c positif *)
let norm_polyf (c,p) =
  let a = content p in
  let b = pgcd (abs_coef a) (abs_coef c) in
  let b = if (lt_coef c coef0) then opp_coef b else b in
    (div_coef c b, div_int p b)
;;

let plusPf (c1,p1) (c2,p2) =
  norm_polyf
    (mult_coef c1 c2,
     plusP (multP (Pint c2) p1) (multP (Pint c1) p2))
;;
let moinsPf (c1,p1) (c2,p2) =
  norm_polyf
    (mult_coef c1 c2,
     moinsP (multP (Pint c2) p1) (multP (Pint c1) p2))
;;

let multPf (c1,p1) (c2,p2) =
  norm_polyf
    (mult_coef c1 c2,
     multP p1 p2)
;;

let oppPf (c,p) =
  norm_polyf (c, oppP p)
;;

(* terme des constructions -> polynome.
   rend un couple (c,p) reprsentant p/c *)
let polyf_of_constr th gl c =
  let rec aux c = 
    match (kind_of_term (strip_outer_cast c)) with 
      | App (binop,[|c1; c2|]) when pf_conv_x gl binop th.th_plus ->
	  plusPf (aux c1) (aux c2)
      | App (binop,[|c1; c2|]) when pf_conv_x gl binop th.th_mult ->
	  multPf (aux c1) (aux c2)
      (* The special case of Zminus *)
      | App (binop, [|c1; c2|])
	  when pf_conv_x gl c
            (mkApp (th.th_plus, [|c1; mkApp(unbox th.th_opp, [|c2|])|])) ->
	    plusPf (aux c1) (oppPf (aux c2))
      | App (unop, [|c1|]) when pf_conv_x gl unop (unbox th.th_opp) ->
	  oppPf (aux c1)
      | _ when pf_conv_x gl c th.th_one ->
	  (coef1, (cf 1))
      | _ when pf_conv_x gl c th.th_zero ->
	  (coef1, (cf 0))
(*      |_ when closed_under th.th_closed c ->
	  mkLApp(coq_SPconst, [|th.th_a; c |])*)
      | _ -> 
	  (try (let n = Hashtbl.find !all_vars c in
                  (coef1,(x n)))
           with _ -> (nvars:=(!nvars)+1;
                      Hashtbl.add !all_vars c !nvars;
		      special_list_vars:=c::!special_list_vars;
                      (*prt(string_of_constr c);*)
                      (coef1,(x !nvars))))
 in aux c



(* Reprsentation des hypothses qui sont des egalites entre polynomes*)
type hpoleq={hname:constr; (* le nom de l'hypothse *)
             hleft:constr;
             hright:constr;
             hpoly:poly; (* hleft-hright *)
            }
;;

(* Transforme une hypothese h:t en equation 
*)
let poleq1_of_constr th gl (h,t) =
  match (kind_of_term t) with
      App (f,args) ->
        (match kind_of_term f with
           | Ind (sp,i) -> 
               (match (string_of_constant sp) with 
		    |"eqT" | "eq" ->
		       let t1= args.(1) in
		       let t2= args.(2) in
			 [{hname=h;
			   hleft=t1;
			   hright=t2;
			   hpoly= snd
				    (moinsPf (polyf_of_constr th gl t1)
				       (polyf_of_constr th gl t2))}
			 ]
		    |_->assert false)
           |_-> assert false)
    |_-> assert false
;;

(* entier -> nombre de Coq *)
let rec int_to_Coq th n =
  if (n<0)
  then (let r = int_to_Coq th (-n) in
          mkApp((unbox th.th_opp), [|r|]))
  else
    if eq_constr th.th_a  (parse ("R")) 
    then  parse ("``"^(string_of_int n)^"``")
    else 
      if  eq_constr th.th_a (parse ("Z")) 
      then parse ("`"^(string_of_int n)^"`")
      else
        parse (string_of_int n) 
;;

let nom_de_variable th n =
  let res=ref th.th_one in
    Hashtbl.iter (fun v i -> if n=i then res:=v) !all_vars;
    !res
;;


(* polynome -> terme des constructions *)
let rec poly_to_constr th p =
  match p with
      Pint a -> int_to_Coq th (int_of_coef a)
    |Prec(v,coefs)->
       let vt=nom_de_variable th v in
       let coefs = Array.map (poly_to_constr th) coefs in
       let n = Array.length coefs in
       let res = ref coefs.(n-1) in
         for i=n-2 downto 0 do
           res:=mkApp(th.th_plus,
                      [|coefs.(i);
                        mkApp(th.th_mult,[|vt;!res|])|]);
         done;
         !res
;;

let rec pol_rec_to_sparse p v1 v2 =  (* v1 et v2 indices de dbut et de fin *)
  let d = v2-v1+1 in
    match p with
        Pint a ->
          if a= coef0
          then []
          else [a, Array.create (d+1) 0]
      |Prec(v,coefs)->
         let vp=Dansideal.gen d (v2-v+1) in (*j'aurais dit v2-v-1*)
         let res=ref [] in
           for i=(Array.length coefs)-1 downto 0 do
             res:= (Dansideal.plusP d
                      (Dansideal.multP d vp (!res))
                      (pol_rec_to_sparse coefs.(i) v1 v2));
           done;
           !res
;;

(* on inverse l'ordre des variables *)       
let rec pol_sparse_to_rec n1 n2 p =
  List.fold_left (fun r (c,m)->
                    let mp = ref (Pint coef1) in
                      Array.iteri (fun i e ->
                                     if i>0 then
                                       mp:=(multP (!mp) (monome (n2-i+1) e)))
                        m;
                      plusP r (multP (Pint c) (!mp)))
    (Pint coef0)
    p
;;  

(******************************************************************************

  La tactique Gb
*)

(* liste de polynomes lpol=p1,...,pn, 
   echoue ou rend q1,...,qn,q tq 1=q1*p1+...+qn*pn+q*(1-zp)

   
   en general (B.Mourrain):
   exprimer 1=q1p1+...qnpn+q(1-zp)
   calculer la base de grobner de 

   t-e,t*(1-zp)-e0 t*p1-e1,...,t*pn-en,ei*ej

   avec t>>e>>x1,...,xm,z=x(m+1)>>e0,e1,...,en
   et recuperer dedans un polynome
   e-q*e0-q1*e1-...-qn*en
   Voila.

*)
let deg_nullstellensatz lpol p =
  (*
  prt "deg_nullstellensatz"; prt "lpol =";
  List.iter (fun p -> prt (string_of_P p)) lpol;
  prt "-----------------";
  prt "p =";
  prt (string_of_P p);
  prt "-----------------";
  *)
  (* lpol = p1, ..., pn *)
  let n = List.length lpol in
  let i = ref 0 in
  let m= (!nvars) in
    (*  t=m+3 e=m+2 ei=-i z=m+1*)
    
  (* lpol1 = t*p1-e1, ..., t*pn-en *)
  let lpol1 = List.map (fun pi ->
                          i:=!i+1;
                          plusP (multP (x (m+3)) pi) (oppP (x (-(!i)))))
                lpol in
    (* polsup = t*(1-z*p) + e0 *)
  let polsup=(plusP (multP (x (m+3)) (plusP cf1 (oppP (multP (x (m+1)) p)))) (oppP (x 0))) in
  let lpol1=polsup::lpol1 in 
    (* t - e *)
  let lpol1=(plusP (x (m+3)) (oppP (x (m+2))))::lpol1 in
  let leiej=ref [] in
   (* ei*ej *)
    for j=0 to n-1 do
      for k=j+1 to n do
        leiej:=(multP (x (-j)) (x (-k)))::(!leiej);
      done;
    done;
    (* e*ei *)
    for j=0 to n do
      leiej:=(multP (x (-j)) (x (m+2)))::(!leiej);
    done;
    let lpol1=(!leiej)@lpol1 in
      (*List.iter (fun p -> print_string ((string_of_P p)^"\n")) lpol1;
      print_string ("----------------------"^"\n");flush(stdout);*)
      let lvar=ref [] in
	for i=0 to n do lvar:=(!lvar)@["e["^(string_of_int i)^"]"]; done;
	for i=1 to m do lvar:=["x["^(string_of_int i)^"]"]@(!lvar); done;
	lvar:=["t";"e";"z"]@(!lvar);
	Dansideal.name_var:=!lvar;
	let lpol1=List.map (fun p -> pol_rec_to_sparse p (-n-1) (m+3)) lpol1 in
(*          Dansideal.lprintP lpol1; print_string
            ("\n----------------------"^"\n");flush(stdout);*)
	  let gb = Dansideal.buch (m+n+4) lpol1 in
          prt "";
(*          print_string ("------------------base calculee---"^"\n");flush(stdout);
          Dansideal.lprintP gb;*)
          (* si on ecrit [q] =.. en fait il faut que q soit de longueur 1 *)
	  let q = (List.filter (fun p -> match p with  (*la*)
                                    [] -> false
				  |(_,mon)::_ -> mon.(2)=1) (*c'est koi ce 2 !!!*)
                     gb) in
            (* si j'ai bien compris, on cherche les polynomes de la base dont les
               monomes de tete contiennent la variable 'e'*)
(*            print_string ("\n-------------poly principal---"^"\n");flush(stdout);
              Dansideal.lprintP q;*)
            (* la base est reduite donc q est de longueur 1 ,
	       soit q' son polynome de tete*)
	  let q'=List.nth q 0 in 
	  let lq = ref [] in
	  let degz = ref 0 in           
            for i=1 to n+1 do
              let i0=n+m+5-i in
		(* on cherche les monomes qui sont construits a
		   l'aide des ei *)
              let qi=List.filter (fun p -> match p with
                                      (_,mon) ->  mon.(i0)>0 
					(* |[] -> false*)) 
                       q' in 
		(*je multiplie les monomes precedents par le coeff de "e" *)
		(*attention je suppose que ce coeff est 1 ou -1*)
		(*je viens de trouver un exemple ou c'est diff de 1 ou -1*)
		(*on peut pas diviser par le coeff car les futurs coeff sont dans Z*)
              let qi = List.map (fun p -> match p with 
                                     (a,mon) -> 
                                       ((mult_coef a 
					   (opp_coef
					      (coef_of_int
						 (signe_coef (fst (List.nth q' 0)))))),mon))
			 qi in
              let qqi = List.map (fun p ->
                                    let mon =snd p in
                                    let c = fst p in
                                    let mon1=Array.copy mon in
                                      mon1.(i0)<-mon.(i0)-1;
                                      (c,mon1)) qi in
		List.iter (fun qi ->
                             let deg' = (snd qi).(3) in
                               (if (compare deg' !degz)=1 then 
				  degz:=deg'))
		  qqi;
		lq:=[qqi]@(!lq);
(*		prt "\nles qqi"; Dansideal.lprintP [qqi];prt "\n";*)
		(*lq:=[qqi]@(!lq);*)
            done;
(*            prt "--------degr de z----------";
            print_int(!degz);
            prt "";*)
            let pi=pol_rec_to_sparse p (-n-1) (m+3) in
              lq := List.map (fun hi ->
				let gi = Dansideal.specialsubstP (m+4) hi 3 !degz in 
				  Dansideal.substP (m+4) gi 3 pi) !lq;
(*              print_string ("\n---------coef calcules---"^"\n");flush(stdout);
	      Dansideal.lprintP (!lq);*)
              let lqr=List.map (fun p->pol_sparse_to_rec (-n-1) (m+3) p) !lq in
(*		print_string
		  ("----------convertis:------------"^"\n");flush(stdout);*)
(*		List.iter (fun p -> print_string ((string_of_P p)^"\n"))
		  lqr; prt""; prt"fin des calculs"; prt"";*)
		let specialcoef = fst (List.nth q' 0) in
		  (specialcoef,(lqr,!degz))
;;                


(*
  rend la tactique qui prouve que p1,...,pn=0 => p=0,
  on a lpoleq=p1,...,pn et goal = p 
*)
let parse_tac_ast   = Pcoq.parse_string Pcoq.Tactic.tactic

let gb_lpoleq th lpoleq goal =
  let lpol = List.map (fun {hpoly=p1}->p1) lpoleq in
  let {hleft=g1;hright=g2;hpoly=p} = goal in  
  let deg_null = deg_nullstellensatz lpol p in 
  let degz = snd (snd (deg_null))  in 
  let lq = List.tl (fst (snd (deg_null))) in
  let specialcoef = abs_coef (fst (deg_null)) in
  let tac = ref tclIDTAC in
    (* transformer le but g1 = g2 en g1 - g2 = 0 *)

  let g = ref th.th_zero in
    if g2 = th.th_zero then
      g:=g1
    else
      g:=mkApp (th.th_plus,[|g1;mkApp (unbox th.th_opp,[|g2|])|]);
    (* g = g1 - g2 *)
    let pseudopower = ref !g in
      for i=1 to (degz-1) do
        pseudopower:=mkApp (th.th_mult,[|!pseudopower;!g|])
      done;
      (* pseudopower = (g * ...) * g degz fois *)
      if g2 <> th.th_zero then
        tac:=(tclTHEN (apply (mkApp (parse "Th_minus_eq",
				     [|th.th_a;
				       th.th_plus;
				       th.th_zero;
				       unbox th.th_opp;
				       g1;
				       g2|]))) 
		!tac);
      (* le but est maintenant g=0 *)
      
(*
	      print_string "\nlq:\n";
	      List.iter (fun q -> print_string (string_of_P q);print_string "\n") lq;
	      print_string "\nlpol:\n";
	      List.iter (fun q -> print_string (string_of_P q);print_string "\n") lpol;
*)
      (* sqp = (q1*p1+...)+qn*pn *)

      let sqp = ref th.th_zero in
        List.iter2 (fun qi pi ->
		      let qipi= mkApp(th.th_mult,
				      [|poly_to_constr th qi;poly_to_constr th pi|]) in
			sqp:=mkApp(th.th_plus,[|!sqp;qipi|]))
          lq lpol;
        
	let tac3 = ref tclIDTAC in
	  (* tranformer les hypotheses a=b en a-b=0 *)
	  (* et reecrire les polynomes des hypotheses en notation de Horner *)
          List.iter
            (fun {hname=h;hleft=a;hright=b;hpoly=p}->
               let hn = match kind_of_term h with 
		   Var n -> n 
		 |_ -> assert false in
               let hp = poly_to_constr th p in
		 tac3:=(tclTHEN !tac3
			  (tclTHENS
			     (replace
				hp th.th_zero)
			     [tclIDTAC;
			      tclTHEN (tclTRY (Ring.polynom[]))
				(tclTHENS 
				   (replace th.th_zero (mkApp (th.th_plus,[|a;mkApp (unbox th.th_opp,[|b|])|])))
				   [tclTRY (Ring.polynom[]); 
				    (tclTHEN (Equality.rewriteLR h) (tclTRY (Ring.polynom[])))])
			     ]));
            ) lpoleq;
	  
          tac3:=tclTHEN !tac3 (tclTRY (Ring.polynom[]));
	  
	  let pseudopower1 = (mkApp (th.th_mult,
				[|poly_to_constr th (Pint  specialcoef);!pseudopower|])) in
          tac:=tclTHEN !tac
	    (tclTHENS (true_cut (Some (id_of_string "Ring_is_a_domain"))
			 (mkApp (parse "is_domain",[|th.th_a;th.th_mult;th.th_zero|])))
	       [(* l' anneau est integre *)
		 tclIDTAC ;
		 (* g=0 *)
		 (tclTHENS (cut_intro (mkApp (build_coq_eqT (),
					      [|th.th_a;
						pseudopower1;
						th.th_zero|])))
		    [(* a*g^r=0 |- g=0 *)
		      tclTHENS (cut_intro (mkApp (build_coq_eqT (),
						 [|th.th_a;
						   !pseudopower;
						   th.th_zero|])))
		       [(* a*g^r=0, g^r=0 |- g=0 *)
			 default_auto;
			 (* a*g^r=0 |- g^r=0 *)
			 apply (mkApp (parse "Th_mult_cte",
				       [|th.th_a;th.th_mult;th.th_zero;
					 mkVar (id_of_string "Ring_is_a_domain");
					 poly_to_constr th (Pint  specialcoef);
					 !pseudopower|]))
		       ];
		      (* a*g^r=0 *)
		      tclTHENS (apply (mkApp (parse "trans_eqT",
					      [|th.th_a;
						pseudopower1;
						!sqp;
						th.th_zero|])))
			[(* a*g^r=q1*p1+...+qn*pn *)
			  Ring.polynom[];
			  (* q1*p1+...+qn*pn = 0 *)
			  !tac3]
		    ])
	       ]);
	  !tac;
;;



let list_of_sign s = List.map (fun (x,_,z)->(x,z)) s;;



let rec gb gl=
  (match Hipattern.match_with_equation (pf_concl gl) with 
     |Some (eq,t::args) ->
        let th = guess_theory t in
        let parse = pf_parse_constr gl in
        let goal = strip_outer_cast (pf_concl gl) in
        let fhyp = id_of_string "new_hyp_for_gb" in
          (* les hypothses *)
        let hyps = List.map (fun (h,t)-> (mkVar h,(body_of_type t)))
                     (list_of_sign (pf_hyps gl)) in
        let lpoleq =ref [] in
          List.iter (fun h -> try (lpoleq:=(poleq1_of_constr th gl h)@(!lpoleq))
                     with _-> ())
            hyps;
          (* lpoleq = les inquations dcoulant des hypothses *)
          let tac = ref tclIDTAC in
          let n = List.length !lpoleq in
            if n = 0 then 
              tac := tclTHEN !tac (tclTRY (Ring.polynom[]))
            else
              tac := tclTHEN !tac (gb_lpoleq th (!lpoleq)
				     (List.hd
					(poleq1_of_constr th gl (mkVar (id_of_string "goal"),goal))));
            Hashtbl.clear !all_vars;
	    special_list_vars:=[];
            nvars:=0;
            (!tac gl);
     |_->(tclIDTAC gl))
;;



TACTIC EXTEND Gb0
      | [ "Gb0" ] -> [ gb]
          END
